/*
    (c) 2002,2003 by Philipp Dunkel philipp@dunkel.org
*/
/*
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
/*
    This program was written as a module for the Apache Webserver (c) by the Apache Foundation
    It is important to note that this implies the use of Libraries and APIs under different
    licenses.
    This means that the GNU General Public License applies only to the Content's of this
    Program, and not necessarily to any used Library.
    On the other hand this also means that that the Apache License (et al.) do not apply to
    the sources of this program or any use thereof.
    
    In Laymens terms: Apache is Apache, mod_xslt is mod_xslt. Obey the Apache Licence for Apache,
        for mod_xslt obey the GNU General Public License.
*/

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "ap_config.h"
#include "http_log.h"
#include "apr_buckets.h"
#include "util_filter.h"
#include "apr_strings.h"
#include "apr_hash.h"
#include "apr_lib.h"
#include "apr_pools.h"
#include "apr_buckets.h"
#include "pcreposix.h"

#include <libxml/parser.h>
#include <libxml/uri.h>
#include <libxslt/xslt.h>
#include <libxslt/transform.h>

#ifndef XSL_LINENUMBERSDEFAULT
#define XSL_LINENUMBERSDEFAULT 1
#endif

/*
 * The Module Structure for the mod_xslt module
 */
module mod_xslt;

/*
 * Configuration Structures
 */
typedef struct {
    apr_pool_t *pool;
    int process;
    int cache;
    char *defstyle;
} dir_config;

/*
 * Stylesheet Cache
 */
apr_hash_t *xsl_cache = NULL;
apr_pool_t *xsl_cache_pool = NULL;

/*
 * XSLT Error Handling / Logging Routine
 */
void xslt_error(void *ctx, const char *msg, ...) {
    va_list args;	
    ap_filter_t *filter=(ap_filter_t*)ctx;
    va_start(args,msg);
        ap_log_rerror(APLOG_MARK,APLOG_ERR, 0, filter->r, apr_pvsprintf(filter->r->pool,msg,args));
    va_end(args);
    return;
}

/*
 * Retrieve the Value of a PI attribute
 */
#define IS_BLANK(val) ((((val)==32) || ((val)==8) || ((val)==10) || ((val)==13))? 1 : 0)
#define NEXT(mover) if(*mover > 0) { mover++; }
#define SKIP_BLANKS(mover) while ((*mover > 0) && IS_BLANK(*mover)) { NEXT(mover); }
#define SKIP_EQUAL(mover) while ((*mover > 0) && (*mover != '=')) { NEXT(mover); }
#define SKIP_QUOTE(mover) quote = *mover; NEXT(mover); while(*mover != quote) { NEXT(mover); }

char *getPIattr(ap_filter_t *f, char *content, const char *name) {
    int namelength = strlen(name);
    char quote = 0;
    char *mover = content;
    char *result = NULL;
    char *retval = NULL;
    while (*mover > 0) {
        SKIP_BLANKS(mover);
        if (strncmp(mover,name,namelength) == 0) {
            mover += namelength;
            if (*mover != '=') {
                SKIP_EQUAL(mover);
                NEXT(mover);
                SKIP_QUOTE(mover);
                NEXT(mover);
                continue;
            }
            NEXT(mover);
            SKIP_BLANKS(mover);
            result = mover + 1;
        
            SKIP_QUOTE(mover);
            *mover = 0;
            
            retval=apr_psprintf(f->r->pool,"%s",result);
            *mover = quote;
            return retval;
        } else {
            SKIP_EQUAL(mover);
            NEXT(mover);
            SKIP_QUOTE(mover);
            NEXT(mover);
        }
    }
}

/*
 * Stylesheet Cache Management
 */
#ifndef NOCACHE
#define xslt_CacheStylesheet(url,style) apr_hash_set(xsl_cache,apr_psprintf(xsl_cache_pool,"%s",url),APR_HASH_KEY_STRING,style);
#define xslt_CachedStylesheet(url) apr_hash_get(xsl_cache,url,APR_HASH_KEY_STRING);
#else
#define xslt_CacheStylesheet(url,style)
#define xslt_CachedStylesheet(url) xsltParseStylesheetFile(url);
#endif

/*
 * Finds the Stylesheet to apply and loads it
 */

xmlChar *xslt_FindStylesheetURL(xmlDocPtr document, ap_filter_t *f) {
    char *mode = NULL;
    char *href = NULL;
    char *xnoproc = NULL;
    char *xagent = NULL;
    char *xaccept = NULL;
    regex_t regex;
    regmatch_t match;
    int doesmatch = 0;
    dir_config* config = NULL;
    xmlNodePtr item = NULL;
    const char *agent = apr_table_get(f->r->headers_in,"User-Agent");
    const char *accept = apr_table_get(f->r->headers_in,"Accept");
    
    if (accept == NULL) { accept = apr_psprintf(f->r->pool,""); }
    if (agent == NULL) { agent = apr_psprintf(f->r->pool,""); }
    
    config = (dir_config *)ap_get_module_config(f->r->per_dir_config, &mod_xslt);
    item = document->doc->children;
    while (item) {
        if (item->type == XML_PI_NODE) {
            if (xmlStrEqual(item->name, BAD_CAST "xml-stylesheet")) {
                mode = getPIattr(f,(char*)item->content, "type");
                href = getPIattr(f,(char*)item->content, "href");
                if (mode && href) {
                    if (xmlStrEqual(mode,BAD_CAST "text/xsl") || xmlStrEqual(mode,BAD_CAST "text/xml")) {
                        return xmlBuildURI(href, document->URL);
                    }
                }
            } else if(xmlStrEqual(item->name, BAD_CAST "xsl-stylesheet")) {
                xagent = getPIattr(f,(char*)item->content, "agent");
                xaccept = getPIattr(f,(char*)item->content, "accept");
                href = getPIattr(f,(char*)item->content, "href");
                xnoproc = getPIattr(f,(char*)item->content, "noprocess");
                if ((xagent || xaccept) && (href || (xnoproc && xmlStrEqual(xnoproc,"yes")))) {
                    doesmatch = 1;
                    if (xagent) {
                        apr_table_add(f->r->headers_out, "Vary", "User-Agent");
                        if (agent) {
                            if (regcomp(&regex,xagent, 0) == 0) {
                                if (regexec(&regex, agent,1,&match,0)) {
                                    doesmatch = 0;
                                }
                                regfree(&regex);
                            } else {
                                doesmatch = 0;
                            }
                        }
                    }
                    if (xaccept) {
                        apr_table_add(f->r->headers_out, "Vary", "Accept");
                        if (accept) {
                            if (regcomp(&regex,xaccept, 0) == 0) {
                                if (regexec(&regex, accept,1,&match,0)) {
                                    doesmatch = 0;
                                }
                                regfree(&regex);
                            } else {
                                doesmatch = 0;
                            }
                        }
                    }
                    if (doesmatch) {
                        if (xnoproc && xmlStrEqual(xnoproc,"yes")) {
                            apr_table_set(f->r->notes, "NoProcess", "true");
                            return NULL;
                        }
                        return xmlBuildURI(href, document->URL);
                    }
                }
            }
            item = item->next;
        } else {
            item = NULL;
        }
    }
    if (config->defstyle != NULL) {
       return xmlBuildURI(config->defstyle, document->URL);
    }
    return NULL;
}

xsltStylesheetPtr xslt_LoadStylesheet(xmlDocPtr document, ap_filter_t *f) {
    xmlChar *URL = xslt_FindStylesheetURL(document,f);
    dir_config* config = NULL;
    xsltStylesheetPtr stylesheet = NULL;
    if (URL == NULL) {
        if (apr_table_get(f->r->notes, "NoProcess") == NULL) {
            ap_log_rerror(APLOG_MARK,APLOG_ERR,0,f->r,"Could not find a Stylesheet at: %s",URL);
        } else {
            apr_table_unset(f->r->notes, "NoProcess");
        }
        return NULL;
    }    
    config = (dir_config *)ap_get_module_config(f->r->per_dir_config, &mod_xslt);
    if (config->cache) {
        stylesheet = xslt_CachedStylesheet(URL);
        if (stylesheet == NULL) {
            stylesheet = xsltParseStylesheetFile(URL);
            xslt_CacheStylesheet(URL,stylesheet);
            if (stylesheet == NULL) {
                ap_log_rerror(APLOG_MARK,APLOG_ERR,0,f->r,"Could not find valid Stylesheet at: %s",URL);
            }
        }
    } else {
        stylesheet = xsltParseStylesheetFile(URL);
        if (stylesheet == NULL) {
            ap_log_rerror(APLOG_MARK,APLOG_ERR,0,f->r,"Could not find valid Stylesheet at: %s",URL);
        }
    }
    if (URL) {
        apr_table_set(f->r->notes,"Stylesheet-URL",URL);
    }
    xmlFree(URL);
    
    return stylesheet;
}

/*
 * XSLT Output Handler
 */
int xslt_OutputBufferWrite(void * context, const char * buffer, int len) {
    ap_filter_t *filter = (ap_filter_t*)context;
    apr_bucket_brigade *brigade = (apr_bucket_brigade *)filter->ctx;

    char *temp = malloc(len);
    if (temp != NULL) {
        memcpy(temp,buffer,len);
        APR_BRIGADE_INSERT_TAIL(brigade, apr_bucket_heap_create(temp,len,&free, filter->r->connection->bucket_alloc));
        return len;
    } else {
        return -1;
    }
}

/*
 * Make Parameter Hash
 */
 
 char *xslt_QuoteParameter(apr_pool_t *pool, char *original) {
    char *mover = original;
    while (*mover) {
        if ((*mover == '\'')||(*mover =='\"')) {
            *mover = '`';
        }
        mover++;
    }
    return apr_psprintf(pool,"'%s'",original);
 }
 
 #define XSLT_STDPARAMS 17
 void xslt_AddStandardParameters(ap_filter_t *f, char **params) {
    params[0] = apr_psprintf(f->r->pool,"%s","HTTP_SERVER_NAME");
    params[1] = apr_psprintf(f->r->pool,"%s",f->r->hostname);
    
    params[2] = apr_psprintf(f->r->pool,"%s","HTTP_SERVER_PORT");
    params[3] = apr_psprintf(f->r->pool,"%d",f->r->server->port?f->r->server->port:80);
    
    params[4] = apr_psprintf(f->r->pool,"HTTP_REQUEST_URL");
    params[5] = apr_psprintf(f->r->pool,"http://%s:%d%s",f->r->hostname,f->r->server->port?f->r->server->port:80,f->r->uri);
    
    params[6] = apr_psprintf(f->r->pool,"%s","HTTP_URI");
    params[7] = apr_psprintf(f->r->pool,"%s",f->r->uri);
    
    params[8] = apr_psprintf(f->r->pool,"%s","HTTP_QUERY_STRING");
    params[9] = apr_psprintf(f->r->pool,"%s",f->r->args);
    
    params[10] = apr_psprintf(f->r->pool,"%s","HTTP_FILENAME");
    params[11] = apr_psprintf(f->r->pool,"%s",f->r->filename);
    
    params[12] = apr_psprintf(f->r->pool,"%s","HTTP_XSLT_STYLESHEET");
    params[13] = apr_psprintf(f->r->pool,"%s",apr_table_get(f->r->notes,"Stylesheet-URL"));
    
    params[14] = apr_psprintf(f->r->pool,"%s","HTTP_ACCEPT");
    params[15] = apr_psprintf(f->r->pool,"%s",apr_table_get(f->r->headers_in,"Accept"));
    
    params[16] = apr_psprintf(f->r->pool,"%s","HTTP_ACCEPT_LANGUAGE");
    params[17] = apr_psprintf(f->r->pool,"%s",apr_table_get(f->r->headers_in,"Accept-Language"));
    
    params[18] = apr_psprintf(f->r->pool,"%s","HTTP_USERAGENT");
    params[19] = apr_psprintf(f->r->pool,"%s",apr_table_get(f->r->headers_in,"User-Agent"));
    
    params[20] = apr_psprintf(f->r->pool,"%s","HTTP_ACCEPT_ENCODING");
    params[21] = apr_psprintf(f->r->pool,"%s",apr_table_get(f->r->headers_in,"Accept-Encoding"));
    
    params[22] = apr_psprintf(f->r->pool,"%s","HTTP_SERVER_ADDRESS");
    params[23] = apr_psprintf(f->r->pool,"%s",f->r->connection->local_ip);
    
    params[24] = apr_psprintf(f->r->pool,"%s","HTTP_METHOD");
    params[25] = apr_psprintf(f->r->pool,"%s",f->r->method);
    
    params[26] = apr_psprintf(f->r->pool,"%s","HTTP_REMOTE_ADDRESS");
    params[27] = apr_psprintf(f->r->pool,"%s",f->r->connection->remote_ip);
    
    params[28] = apr_psprintf(f->r->pool,"%s","HTTP_SERVER_ADMIN");
    params[29] = apr_psprintf(f->r->pool,"%s",f->r->server->server_admin);
    
    params[30] = apr_psprintf(f->r->pool,"%s","HTTP_UPARSED_URI");
    params[31] = apr_psprintf(f->r->pool,"%s",f->r->unparsed_uri);
    
    params[32] = apr_psprintf(f->r->pool,"%s","HTTP_REMOTE_USER");
    params[33] = apr_psprintf(f->r->pool,"%s",f->r->user?f->r->user:"");
 }
 
 char ** xslt_MakeParameterHash(ap_filter_t *f) {
    char hexbuf[3];
    char **result = NULL;
    char **resmov = NULL;
    char *querystr= apr_psprintf(f->r->pool,"%s",f->r->args?f->r->args:"");
    char *mover = querystr;
    int paramcnt = 0;
    while (*mover > 0) {
        if (*mover == '&') {
            paramcnt++;
        }
        mover++;
    }
    paramcnt++;
    result = apr_palloc(f->r->pool,(sizeof(char*) * (XSLT_STDPARAMS + paramcnt + 1) * 2));
    resmov = result;
    while (resmov < (result + ((XSLT_STDPARAMS + paramcnt + 1) * 2))) {
        *resmov = NULL;
        resmov++;
    }
    mover = querystr;
    resmov = result;
    resmov[0] = mover;
    resmov[1] = "";
    while (*mover > 0) {
        if ((*mover == '&') || (*mover == '=')) {
            *mover = 0;
            resmov++;
            *resmov = mover + 1;
        }
        mover ++;
    }
    resmov = result;
    while (*resmov) {
        mover = *resmov;
        while (*mover > 0) {
            if (*mover == '%') {
                strncpy(hexbuf,mover+1,2);
                *mover = (char)strtol(hexbuf,NULL,16);
                *(mover + 1) = 0;
                strcat(*resmov,mover+3);
            }
            mover++;
        }
        resmov++;
    }
    xslt_AddStandardParameters(f,resmov);
    resmov = result;
    while (*resmov) {
        resmov++;
        *resmov = xslt_QuoteParameter(f->r->pool,*resmov);
        resmov++;
    }
    return result;
}

/*
 * Utility for getting XML from a Bucket-Brigade
 */
xmlDocPtr xslt_ParseBrigadeDocument(ap_filter_t* f, apr_bucket_brigade *brigade) {
    xmlDocPtr document = NULL;
    char *buffer = NULL;
    apr_off_t length = 0;
    apr_pool_t *pool = NULL;
    apr_pool_create(&pool,f->r->pool);
    apr_brigade_length(brigade,1,&length);
    apr_brigade_pflatten(brigade,&buffer,(apr_size_t *)&length,pool);
    xmlSetGenericErrorFunc((void*)f, xslt_error);
    xsltSetGenericErrorFunc((void*)f, xslt_error);
    document = xmlParseMemory(buffer, length);
    xmlSetGenericErrorFunc(NULL, NULL);
    xsltSetGenericErrorFunc(NULL, NULL);
    apr_pool_destroy(pool);
    return document;
}

/*
 * mod_xslt Transformation Routine
 */
static apr_status_t xslt_parse(ap_filter_t *f,apr_bucket_brigade *brigade) {
    char *fbase = NULL;
    char *hbase = NULL;
    xsltStylesheetPtr stylesheet = NULL;
    int ret = 0;
    
    xmlDocPtr document = xslt_ParseBrigadeDocument(f, brigade);

    xmlOutputBufferPtr output;
    xmlCharEncodingHandlerPtr encoder = NULL;

    char **parameters = NULL;
    
    dir_config* config = (dir_config *)ap_get_module_config(f->r->per_dir_config, &mod_xslt);
    
    apr_bucket_brigade *resultbrigade = apr_brigade_create(f->r->pool,NULL);

    if (!document) { return ap_pass_brigade(f->next, brigade); }
    
    document->URL=xmlBuildURI(f->r->uri, apr_psprintf(f->r->pool,"http://%s:%d%s",f->r->hostname,f->r->server->port?f->r->server->port:80,f->r->uri));
    
    stylesheet = xslt_LoadStylesheet(document, f);
    if (stylesheet == NULL) {
        xmlFreeDoc(document);
        return ap_pass_brigade(f->next, brigade); 
    }

    apr_table_unset(f->r->headers_out, "Content-Type");
    apr_table_unset(f->r->headers_out, "Content-Length");
    apr_table_unset(f->r->headers_out, "Last-Modified");

    if (stylesheet->mediaType) {
        apr_table_set(f->r->notes,"Media-Type",stylesheet->mediaType);
    } else if (stylesheet->method) {
        if (!strcasecmp(stylesheet->method,"html")) {
            apr_table_set(f->r->notes,"Media-Type","text/html");
        } else if (!strcasecmp(stylesheet->method,"xml")) {
            apr_table_set(f->r->notes,"Media-Type","text/xml");
        } else if (!strcasecmp(stylesheet->method,"text")) {
            apr_table_set(f->r->notes,"Media-Type","text/plain");
        }
    } else {
        apr_table_set(f->r->notes,"Media-Type","text/html");
    }
    if (stylesheet->encoding) {
        apr_table_set(f->r->notes,"Charset",stylesheet->encoding);
    } else {
        apr_table_set(f->r->notes,"Charset","UTF-8");
    }
        
    encoder = xmlFindCharEncodingHandler((const char*)apr_table_get(f->r->notes,"Charset"));
    if (encoder && ((encoder->input == NULL) || (encoder->output == NULL))) {
        encoder=NULL;
        apr_table_unset(f->r->notes, "Charset");
    }

    if (f->r->header_only) {
        xmlFreeDoc(document);
        if (!config->cache) {
            xsltFreeStylesheet(stylesheet);
        }
    } else {
        output = xmlAllocOutputBuffer(encoder);
        if (!output) {
            xmlFreeDoc(document);
            if (!config->cache) {
                xsltFreeStylesheet(stylesheet);
            }
            return ap_pass_brigade(f->next, brigade);
        }
    
        f->ctx = resultbrigade;
        output->context = (void*)f;
    
        output->writecallback = xslt_OutputBufferWrite;
        output->closecallback = NULL;
        
        parameters = xslt_MakeParameterHash(f);
        
        xmlSetGenericErrorFunc((void*)f, xslt_error);
        xsltSetGenericErrorFunc((void*)f, xslt_error);    
        ret = xsltRunStylesheet(stylesheet, document, (const char**)parameters, NULL, NULL, output);
        xmlSetGenericErrorFunc(NULL, NULL);
        xsltSetGenericErrorFunc(NULL, NULL);        
    
        xmlFreeDoc(document);
        if (!config->cache) {
            xsltFreeStylesheet(stylesheet);
        }
        APR_BRIGADE_INSERT_TAIL(resultbrigade, apr_bucket_eos_create(f->r->connection->bucket_alloc));
        xmlOutputBufferClose(output);
    
        if (ret < 1) {
            return ap_pass_brigade(f->next, brigade);
        }
    }
    
    if (apr_table_get(f->r->notes,"Charset") == NULL) {
        apr_table_set(f->r->headers_out,"Content-Type",apr_table_get(f->r->notes,"Media-Type"));
    } else {
        apr_table_set(f->r->headers_out, "Content-Type", apr_psprintf(f->r->pool,"%s; charset=%s",apr_table_get(f->r->notes,"Media-Type"),apr_table_get(f->r->notes,"Charset")));
    }
    f->r->content_type=apr_table_get(f->r->headers_out, "Content-Type");            
    
    return ap_pass_brigade(f->next, resultbrigade);
}

/*
 * The actual mod_xslt filtering routine
 */
#define isfinalbrigade(brigade) APR_BUCKET_IS_EOS(APR_BRIGADE_LAST(brigade))
static apr_status_t xslt_filter(ap_filter_t *f, apr_bucket_brigade *brigade) {
    char *inbuf = NULL;
    int lenght = 0;
    apr_status_t result;
    apr_bucket_brigade *storeddata;

    dir_config *config = (dir_config *)ap_get_module_config(f->r->per_dir_config, &mod_xslt);
    ap_log_rerror(APLOG_MARK,APLOG_DEBUG, 0, f->r, "config->process = %d",config->process);
    ap_log_rerror(APLOG_MARK,APLOG_DEBUG, 0, f->r, "config->cache = %d",config->cache);
    ap_log_rerror(APLOG_MARK,APLOG_DEBUG, 0, f->r, "config->defstyle = %s",config->defstyle);
    if (config->process == 0) {
        return ap_pass_brigade(f->next, brigade);
    }

    ap_log_rerror(APLOG_MARK,APLOG_DEBUG, 0, f->r, "Processing: %s \t\t START",f->r->filename);
    if (f->ctx) {
        storeddata=(apr_bucket_brigade*)f->ctx;
    } else { 
        storeddata=NULL;
    }
    if (isfinalbrigade(brigade)) {
        if (storeddata) {
            APR_BRIGADE_CONCAT(brigade, storeddata);
        }
        result = xslt_parse(f, brigade);
    } else {
        ap_save_brigade(f, &storeddata, &brigade, f->r->pool);
        f->ctx=storeddata;
        result = APR_SUCCESS;
    }
    ap_log_rerror(APLOG_MARK,APLOG_DEBUG, 0, f->r, "Processing: %s \t\t END",f->r->filename);
    return result;
}

/*
 * Per Process Initialisation Routines
 */
static apr_status_t xslt_child_exit(void *data) {
        char *hashkey = NULL;
        xsltStylesheetPtr style = NULL;
        
#ifndef NOCACHE
        apr_hash_index_t *idx = apr_hash_first(xsl_cache_pool,xsl_cache);
        while (idx) {
            apr_hash_this(idx,NULL,0,(void**)&style);
            xsltFreeStylesheet(style);
            apr_hash_set(xsl_cache,hashkey,APR_HASH_KEY_STRING, NULL);
            idx = apr_hash_next(idx);
        }
#endif
	xsltCleanupGlobals();
	xmlCleanupParser();
	xmlMemoryDump();
	return APR_SUCCESS;
}

static void xslt_child_init(apr_pool_t *p, server_rec *s) {
	ap_log_perror(APLOG_MARK,APLOG_DEBUG, 0, p, "Initializing Child");	
	apr_pool_cleanup_register(p, p, xslt_child_exit, xslt_child_exit);
	xmlInitMemory();
	xmlLineNumbersDefault(XSL_LINENUMBERSDEFAULT);
	xmlSubstituteEntitiesDefault(1);
	xmlLoadExtDtdDefaultValue=1;
        xmlDefaultSAXHandlerInit();
        xmlDefaultSAXHandler.cdataBlock = NULL;
#ifndef NOCACHE
        xsl_cache = apr_hash_make(p);
        xsl_cache_pool = p;
#endif
}

/*
 * mod_xslt Hook Registration
 */
static void xslt_register_hooks(apr_pool_t *p) {
	ap_hook_child_init(xslt_child_init, NULL, NULL, APR_HOOK_MIDDLE);
	ap_register_output_filter("mod_xslt", xslt_filter, NULL, AP_FTYPE_RESOURCE);
}

/*
 * mod_xslt Configuration Management Routines
 */

static void *xslt_create_dir_config(apr_pool_t *p, char *dir) {
	dir_config* nconf=(dir_config*)apr_palloc(p,sizeof(dir_config));
        nconf->pool=p;
	nconf->cache=1;
        nconf->process=1;
        nconf->defstyle = NULL;
	return nconf;
}

static void *xslt_merge_dir_config(apr_pool_t *p, void *basev, void *overridesv) {
        dir_config* override = (dir_config*)overridesv;
	dir_config* nconf = (dir_config*)apr_palloc(p,sizeof(dir_config));
        nconf->pool=p;
	nconf->cache=override->cache;
        nconf->process=override->process;
        nconf->defstyle=override->defstyle;
	return nconf;
}

/*
 * mod_xslt Configuration Command Handlers
 */
const char *xslt_XSLTDefault(cmd_parms *cmd, void *pcfg, const char *value) {
	dir_config* conf = (dir_config*) pcfg;
	if (strcmp(value,"off") == 0) {
            conf->defstyle = NULL;
        } else {
            conf->defstyle = apr_psprintf(conf->pool,"%s",value);
        }
	return NULL;
}

const char *xslt_XSLTCache(cmd_parms *cmd, void *pcfg, const char *value) {
	dir_config* conf = (dir_config*)pcfg;
        if (strcmp(value,"on") == 0) {
		conf->cache=1;
	} else if (strcmp(value,"off") == 0) {
		conf->cache=0;
	} else {
            return "XSLTCache < on | off >";
        }
	return NULL;
}

const char *xslt_XSLTProcess(cmd_parms *cmd, void *pcfg, const char *value) {
	dir_config* conf = (dir_config*)pcfg;
	if (strcmp(value,"on") == 0) {
		conf->process=1;
	} else if (strcmp(value,"off") == 0) {
		conf->process=0;
	} else {
            return "XSLTProcess < on | off >";
        }
	return NULL;
}

/*
 * Definition of configuration commands
 */
static const command_rec xslt_cmds[] = {
        AP_INIT_TAKE1("XSLTProcess", xslt_XSLTProcess, NULL, OR_FILEINFO, "XSLTProcess < on | off >"),
	AP_INIT_TAKE1("XSLTDefault", xslt_XSLTDefault, NULL, OR_FILEINFO, "XSLTDefault < Stylesheet-URL | off >"),
	AP_INIT_TAKE1("XSLTCache", xslt_XSLTCache, NULL, OR_FILEINFO, "XSLTCache < on | off >"),
	{NULL}
};

/*
 * mod_xslt Module definition
 */
module AP_MODULE_DECLARE_DATA mod_xslt = {
	STANDARD20_MODULE_STUFF, 
	xslt_create_dir_config,     /* create per-dir    config structures */
	xslt_merge_dir_config,      /* merge  per-dir    config structures */
	NULL,                       /* create per-server config structures */
	NULL,                       /* merge  per-server config structures */
	xslt_cmds,                  /* table of config file commands       */
	xslt_register_hooks         /* register hooks                      */
};
